#ifndef NTRU_ERR_H
#define NTRU_ERR_H

#define NTRU_SUCCESS 0
#define NTRU_ERR_OUT_OF_MEMORY 1
#define NTRU_ERR_PRNG 2
#define NTRU_ERR_MSG_TOO_LONG 3
#define NTRU_ERR_INVALID_MAX_LEN 4
#define NTRU_ERR_DM0_VIOLATION 5
#define NTRU_ERR_NO_ZERO_PAD 6
#define NTRU_ERR_INVALID_ENCODING 7
#define NTRU_ERR_NULL_ARG 8
#define NTRU_ERR_UNKNOWN_PARAM_SET 9
#define NTRU_ERR_INVALID_PARAM 10
#define NTRU_ERR_INVALID_KEY 11

#endif   /* NTRU_ERR_H */
