#ifndef NTRU_ARITH_H
#define NTRU_ARITH_H

#include <stdint.h>

/**
 * @brief logarithm
 *
 * Returns floor(log(n)).
 *
 * @param n
 * @return
 */
uint8_t ntru_log2(uint16_t n);

#endif   /* NTRU_ARITH_H */
